import scala.language.higherKinds

trait Monad[M[*]]:
  def pure[A](a: A): M[A]
  def bind[A,B](m: M[A], f: A => M[B]): M[B]

object Monad:
  implicit class MonadOps[M[*], A](m: M[A])(implicit instance: Monad[M]):
    def flatMap[B](f: A => M[B]): M[B] = instance.bind(m, f)
    def map[B](f: A => B): M[B] = flatMap(f andThen instance.pure)  

enum Failable[A]:
  case Success(value: A)
  case Failure(e: Throwable)

  def success[A](value: A): Failable[A] = Success(value)
  def fail[A](e: Throwable): Failable[A] = Failure[A](e)

  implicit object MonadInstance extends Monad[Failable]:
    def pure[A](a: A): Failable[A] = Success(a)
    def bind[A, B](m: Failable[A], f: A => Failable[B]): Failable[B] =
      m match
        case Success(a) => f(a)
        case Failure(e) => Failure(e)      

  import Monad._

  def example = for
    a <- success(4)
    b <- fail[Int](new Exception("whaaat?"))
    c <- success(7)
  yield c

case class StateT[M[*],S,A](run: S => M[(A,S)])

object StateT:
  implicit def monadInstance[M[*],S](implicit m: Monad[M]) =
    new Monad[({type λ[Y] = StateT[M,S,Y]})#λ]:
      override def pure[A](a: A): StateT[M, S, A] =
        StateT(s => m.pure((a,s)))

      override def bind[A, B](f: StateT[M, S, A], g: A => StateT[M, S, B]): StateT[M, S, B] =
        StateT[M,S,B](s =>
          m.bind[(A,S),(B,S)](f.run(s), { case (a,s2) => g(a).run(s2) }))    

  def get[M[*],S](implicit m: Monad[M]): StateT[M,S,S] = new StateT(s => m.pure(s,s))
  def put[M[*],S](s_ : S)(implicit m: Monad[M]): StateT[M,S,Unit] = new StateT(s => m.pure((),s_))

  implicit def lift[M[*],S,A](v: M[A])(implicit m: Monad[M]): StateT[M,S,A] =
    StateT(s => m.bind(v,(a: A) => m.pure((a,s))))

  import Monad._

  def example = for
    a <- get[Failable,Int]
  yield a